import _pydmc


def test__mhash3():
    h1 = _pydmc.hash('foo')
    h2 = _pydmc.hash('foo', h1 & 0xFFFFFFFF)
    print h1, h2
    assert (-39287385592190013122878999397579195001,
            -73964642705803263641983394469427790275) == (h1, h2)


def test_null_key():
    h0 = _pydmc.hash('foo')
    h1 = _pydmc.hash('foo\0bar')
    h2 = _pydmc.hash('foo\0baz')
    assert h0 != h1, 'Hash collision for appended null'
    assert h0 != h2, 'Hash collision for appended null'
    assert h1 != h2, 'Hash collision for bytes after null'


def test_mhash3_long():
    h1 = _pydmc.hash_long(123)
    h2 = _pydmc.hash_long(123, h1 & 0xFFFFFFF)

    assert (-121703982708902402444108248539236701464,
            30126007557438804793814493095132085929) == (h1, h2)



from array import array
import _pydmc


def test_pydmc():
    dmc = _pydmc.StringDMC(500)
    dmc.add('test')

    assert 'test' in dmc
    assert 'duck' not in dmc


def test_long_pydmc():
    dmc = _pydmc.LongDMC(500)

    assert dmc.items == 0
    dmc.add(1015L)
    assert dmc.items == 1
    dmc.add(1015L)
    assert dmc.items == 1

    assert dmc.size == 500

    assert 1015L in dmc
    assert 1015 in dmc
    assert 2015 not in dmc

    origin = dmc.to_byte_array().tostring().encode('hex')
    dmc_2 = _pydmc.LongDMC.from_byte_array(dmc.to_byte_array())
    clone = dmc_2.to_byte_array().tostring().encode('hex')
    assert origin == clone

    assert 1015L in dmc_2
    assert 1015 in dmc_2
    assert 2015 not in dmc_2

    assert dmc.size == dmc_2.size
    assert dmc.items == dmc_2.items

    base64 = dmc.to_base64()
    dmc_3 = _pydmc.LongDMC.from_base64(base64)

    assert 1015L in dmc_3
    assert 1015 in dmc_3
    assert 2015 not in dmc_3

    assert dmc.size == dmc_3.size
    assert dmc.items == dmc_3.items

    dmc.clear()

    assert 1015 not in dmc
    assert dmc.size == 500
    assert dmc.items == 0


def test_extend():
    dmc = _pydmc.LongDMC(500)

    dmc.extend([10, 25, 35])

    dmc2 = _pydmc.LongDMC(500)
    dmc2.extend_array(array('l', [10, 25, 35]))

    assert dmc2.to_base64() == dmc.to_base64()


def test_extend_uint():
    dmc = _pydmc.UIntDMC(500)

    dmc.extend([10, 25, 35])

    dmc2 = _pydmc.UIntDMC(500)
    dmc2.extend_array(array('I', [10, 25, 35]))

    assert dmc2.to_base64() == dmc.to_base64()



from setuptools import setup
from setuptools.extension import Extension
from Cython.Distutils import build_ext
from os.path import join
import functools

import os

rel = functools.partial(join, os.getcwd())

ext_modules = [
    Extension(
        "_pydmc",
        extra_compile_args=['-std=gnu99', '-O2', '-D_LARGEFILE64_SOURCE'],
        sources=["src/_pydmc.pyx",
                 'src/MurmurHash3.c'],

        include_dirs=[rel('src')],
        library_dirs=[rel('src')]
    )
]

setup(
    name='pydmc',
    version='0.2',
    author='Jake Heinz',
    author_email='me@jh.gg',
    url="http://github.com/jhgg/pydmc",
    description='A high performance python direct-memory-cache type thing.',
    license='MIT License',
    cmdclass={'build_ext': build_ext},
    zip_safe=False,
    package_dir={'': 'src'},
    py_modules=['pydmc'],
    ext_modules=ext_modules,
    test_suite='nose.collector'
)



from cpython cimport array
cimport cpython

import bitarray
import struct
import base64

from libc.string cimport memcpy
cdef array.array char_array_template = array.array('b', [])

cdef extern from "MurmurHash3.h" nogil:
    void MurmurHash3_x64_128(void *key, int len, unsigned int seed, void *out)
    void MurmurHash3_x64_128_long(long key, unsigned int seed, void *out)

def hash(key, int seed=0):
    cdef long long result[2]
    if isinstance(key, unicode):
        key = key.encode('utf8')

    MurmurHash3_x64_128(<char*> key, len(key), seed, result)
    return long(result[0]) << 64 | (long(result[1]) & 0xFFFFFFFFFFFFFFFF)

def hash_long(long long key, int seed=0):
    cdef long long result[2]
    MurmurHash3_x64_128_long(key, seed, &result)
    return long(result[0]) << 64 | (long(result[1]) & 0xFFFFFFFFFFFFFFFF)

cdef char* fmt = '!III'
cdef ssize_t header_size = sizeof(unsigned int) * 3

cdef class _DMC:
    cdef unsigned int _size
    cdef unsigned int _items
    cdef object _bitarray

    def __cinit__(self, unsigned long size, cpython.bool _clear=True, unsigned int _items=0):
        self._size = size
        self._items = _items
        self._bitarray = bitarray.bitarray(size)

        if _clear:
            self._bitarray.setall(False)

    cpdef _from_byte_array(self, array.array byte_array):
        (address, size, endianness, unused, allocated) = self._bitarray.buffer_info()
        memcpy(cpython.PyLong_AsVoidPtr(address), byte_array.data.as_chars + header_size,
               byte_array.ob_size - header_size)

    @classmethod
    def from_byte_array(cls, array.array byte_array):
        assert byte_array.ob_size > header_size
        array_size, bit_size, items = struct.unpack_from(fmt, byte_array)
        assert bit_size / 8 <= array_size
        assert array_size == byte_array.ob_size - header_size
        cdef dmc = cls(bit_size, _clear=False, _items=items)
        dmc._from_byte_array(byte_array)
        return dmc

    def to_byte_array(self):
        (address, size, endianness, unused, allocated) = self._bitarray.buffer_info()
        cdef unsigned int length = size + header_size
        cdef array.array byte_array = array.clone(char_array_template, length, False)
        struct.pack_into(fmt, byte_array, 0, size, self._size, self._items)
        memcpy(byte_array.data.as_chars + header_size, cpython.PyLong_AsVoidPtr(address), size)
        return byte_array

    def to_base64(self):
        return base64.b64encode(self.to_byte_array())

    @classmethod
    def from_base64(cls, bytes s):
        return cls.from_byte_array(array.array('b', base64.b64decode(s)))

    property items:
        def __get__(self):
            return self._items

    property size:
        def __get__(self):
            return self._size

    cpdef clear(self):
        self._items = 0
        self._bitarray.setall(False)

cdef class LongDMC(_DMC):
    cpdef add(self, long long item):
        hash_val = hash_long(item) % self._size
        old_val = self._bitarray[hash_val]
        if not old_val:
            self._items += 1
            self._bitarray[hash_val] = True

    cpdef extend(self, items):
        for item in items:
            hash_val = hash_long(<long long>item) % self._size
            old_val = self._bitarray[hash_val]
            if not old_val:
                self._items += 1
                self._bitarray[hash_val] = True

    cpdef extend_array(self, long long[:] items):
        for item in items:
            hash_val = hash_long(item) % self._size
            old_val = self._bitarray[hash_val]
            if not old_val:
                self._items += 1
                self._bitarray[hash_val] = True

    def __contains__(self, long long item):
        hash_val = hash_long(item) % self._size
        return self._bitarray[hash_val]

cdef class UIntDMC(_DMC):
    cpdef add(self, unsigned int item):
        hash_val = hash_long(item) % self._size
        old_val = self._bitarray[hash_val]
        if not old_val:
            self._items += 1
            self._bitarray[hash_val] = True

    cpdef extend(self, items):
        for item in items:
            hash_val = hash_long(<unsigned int>item) % self._size
            old_val = self._bitarray[hash_val]
            if not old_val:
                self._items += 1
                self._bitarray[hash_val] = True

    cpdef extend_array(self, unsigned int[:] items):
        for item in items:
            hash_val = hash_long(item) % self._size
            old_val = self._bitarray[hash_val]
            if not old_val:
                self._items += 1
                self._bitarray[hash_val] = True

    def __contains__(self, unsigned int item):
        hash_val = hash_long(item) % self._size
        return self._bitarray[hash_val]

cdef class StringDMC(_DMC):
    cpdef add(self, item):
        hash_val = hash(item) % self._size
        old_val = self._bitarray[hash_val]
        if not old_val:
            self._items += 1
            self._bitarray[hash_val] = True

    def __contains__(self, item):
        hash_val = hash(item) % self._size
        return self._bitarray[hash_val]
